"""
@project: djangoModel->tool
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis: CURD 工具
@created_time: 2022/9/15 14:14
"""
from pathlib import Path

from django.core.paginator import Paginator, EmptyPage
from django.db.models import F

from main.settings import BASE_DIR
from xj_finance.services.finance_transacts_service import FinanceTransactsService
from ..models import Enroll, EnrollRecord, EnrollSubitem
from ..service.enroll_record_serivce import EnrollRecordServices
from ..service.enroll_subitem_record_service import EnrollSubitemRecordService
from ..service.subitem_extend_service import output_convert
from ..utils.custom_tool import format_params_handle, write_to_log, filter_fields_handler, force_transform_type
from ..utils.j_config import JConfig
from ..utils.j_dict import JDict

module_root = str(Path(__file__).resolve().parent)
# 配置之对象
main_config_dict = JDict(JConfig.get_section(path=str(BASE_DIR) + "/config.ini", section="xj_finance"))
module_config_dict = JDict(JConfig.get_section(path=str(BASE_DIR) + "/config.ini", section="xj_finance"))

sand_box_meet = main_config_dict.sand_box_meet or module_config_dict.sand_box_meet or ""
sand_box_receivable = main_config_dict.sand_box_receivable or module_config_dict.sand_box_receivable or ""


class EnrollServices:
    enroll_fields = [i.name for i in Enroll._meta.fields]

    def __init__(self):
        pass

    @staticmethod
    def enroll_add(params: dict = None):
        """
        报名表新增
        :param params: 添加参数
        :return: data,err
        """
        params, is_pass = force_transform_type(variable=params, var_type="dict")
        # 请求限制类型
        try:
            params = format_params_handle(
                param_dict=params,
                is_validate_type=True,
                filter_filed_list=[
                    "thread_id|int",
                    "category_id|int",
                    "user_id|int",
                    "trading_relate|int",
                    "region_code",
                    "occupy_room",
                    "enroll_status_code|int",
                    "min_number|int",
                    "max_number|int",
                    "min_count_apiece|int",
                    "max_count_apiece|int",
                    "enroll_rule_group_id|int",
                    "price|float",
                    "count|int",
                    "unit",
                    "fee|float",
                    "reduction|float",
                    "subitems_amount|float",
                    "amount|float",
                    "paid_amount|float",
                    "unpaid_amount|float",
                    "commision|float",
                    "deposit|float",
                    "hide_price|int",
                    "hide_user|int",
                    "has_repeat|int",
                    "has_subitem|int",
                    "has_audit|int",
                    "need_vouch|int",
                    "need_deposit|int",
                    "need_imprest|int",
                    "enable_pool|int",
                    "pool_limit|int",
                    "pool_stopwatch|int",
                    "open_time|date",
                    "close_time|date",
                    "launch_time|date",
                    "finish_time|date",
                    "spend_time|date",
                    "create_time|date",
                    "update_time|date",
                    "snapshot|dict",
                    "remark",
                    "finance_invoicing_code"
                ],
            )
        except ValueError as e:
            return None, str(e)

        # 必填参数校验
        must_keys = ["thread_id", "category_id", "user_id"]
        for i in must_keys:
            if not params.get(i, None):
                return None, str(i) + "必填"

        #  ORM添加插入操作
        try:
            instance = Enroll.objects.create(**params)
        except Exception as e:
            return None, str(e)

        return instance.to_json(), None

    @staticmethod
    def enroll_edit(params: dict = None, enroll_id=None, search_param: dict = None, **kwargs):
        """
        报名编辑
        :param params: 修改的参数
        :param enroll_id: 需要修改的报名主键
        :param search_param: 搜索参数, 服务层根据信息等其他搜索字段检索到需要修改的数据
        :return: data, err
        """
        # 空值检查
        params, is_pass = force_transform_type(variable=params, var_type="dict", default={})
        enroll_id, is_pass = force_transform_type(variable=enroll_id, var_type="int")
        search_param, is_pass = force_transform_type(variable=search_param, var_type="dict", default={})
        if not enroll_id and not search_param:
            return None, "无法找到要修改数据，请检查参数"
        # 搜索字段过滤
        if search_param:
            search_param = format_params_handle(
                param_dict=search_param,
                filter_filed_list=[
                    "enroll_id_list|list", "thread_id|int", "thread_id_list|list", "category_id|int", "category_id_list|list", "trading_relate", "region_code",
                    "enroll_status_code|int", "enroll_rule_group_id|int",
                ],
                alias_dict={"enroll_id_list": "id__in", "thread_id_list": "thread_id__in"}
            )

        # 修改内容检查处理
        try:
            params = format_params_handle(
                param_dict=params,
                is_validate_type=True,
                filter_filed_list=[
                    "thread_id|int",
                    "category_id|int",
                    "user_id|int",
                    "trading_relate|int",
                    "region_code",
                    "occupy_room",
                    "enroll_status_code|int",
                    "min_number|int",
                    "max_number|int",
                    "min_count_apiece|int",
                    "max_count_apiece|int",
                    "enroll_rule_group_id|int",
                    "price|float",
                    "count|int",
                    "unit",
                    "fee|float",
                    "reduction|float",
                    "subitems_amount|float",
                    "amount|float",
                    "paid_amount|float",
                    "unpaid_amount|float",
                    "commision|float",
                    "deposit|float",
                    "hide_price|int",
                    "hide_user|int",
                    "has_repeat|int",
                    "has_subitem|int",
                    "has_audit|int",
                    "need_vouch|int",
                    "need_deposit|int",
                    "need_imprest|int",
                    "enable_pool|int",
                    "pool_limit|int",
                    "pool_stopwatch|int",
                    "open_time|date",
                    "close_time|date",
                    "launch_time|date",
                    "finish_time|date",
                    "spend_time|date",
                    "create_time|date",
                    "update_time|date",
                    "snapshot|dict",
                    "remark",
                    "finance_invoicing_code"
                ],
            )
        except ValueError as e:
            return None, str(e)
        if not params:
            return None, "没有可修改的内容"

        # 必填参数校验
        must_keys = ["thread_id", "category_id", "user_id"]
        for i in must_keys:
            if not params.get(i, None):
                return None, str(i) + "必填"

        # 构建ORM，检查是否存在可修改项目
        enroll_obj = Enroll.objects
        if enroll_id:
            enroll_obj = enroll_obj.filter(id=enroll_id)
        elif search_param:
            enroll_obj = enroll_obj.filter(**search_param)

        update_total = enroll_obj.count()
        if update_total == 0:
            return None, "没有找到可修改项目"

        # IO 操作
        try:
            enroll_obj.update(**params)
        except Exception as e:
            return None, "修改异常:" + str(e)
        return enroll_obj.first().to_json(), None

    @staticmethod
    def enroll_delete(enroll_id):
        enroll_obj = Enroll.objects.filter(id=enroll_id)
        if not enroll_obj:
            return None, None
        try:
            enroll_obj.delete()
        except Exception as e:
            return None, "删除异常:" + str(e)
        return None, None

    @staticmethod
    def enroll_detail(enroll_id: int = None, user_id: int = None, simple_return: bool = False):
        """
        报名详情
        :param enroll_id: 报名ID
        :param user_id: 用户ID
        :param simple_return: 是否仅返回基础信息
        :return: data , err
        """
        enroll_id, is_pass = force_transform_type(variable=enroll_id, var_type="int")
        user_id, is_pass = force_transform_type(variable=user_id, var_type="int")
        if not enroll_id:
            return None, "不是有效的报名ID"

        enroll_obj = Enroll.objects.filter(id=enroll_id).extra(select={
            'open_time': 'DATE_FORMAT(open_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            "close_time": 'DATE_FORMAT(close_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            "launch_time": 'DATE_FORMAT(launch_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            "finish_time": 'DATE_FORMAT(finish_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            "spend_time": 'DATE_FORMAT(spend_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            "create_time": 'DATE_FORMAT(create_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            "update_time": 'DATE_FORMAT(update_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
        })
        if not enroll_obj:
            return None, "找不到该报名信息"
        enroll_detail = enroll_obj.first().to_json()

        if not simple_return:
            # 当前用户是否存在报名记录/当前用户是否存在报名分项记录
            enroll_detail["this_user_has_record"] = 0
            enroll_detail["this_user_has_subitem_record"] = 0
            # ================= 获取需要拼接的数据，报名分项，报名记录，报名分项记录 ==========================
            enroll_subitems_list = output_convert(list(EnrollSubitem.objects.annotate(category_id=F("enroll__category_id")).filter(enroll_id=enroll_id).values()))
            main_record_list, err = EnrollRecordServices.record_list({"enroll_id": enroll_id}, need_pagination=False)
            print("main_record_list:", len(main_record_list))
            enroll_subitems_record_list, err = EnrollSubitemRecordService.list({"enroll_id": enroll_id}, False)
            # ================= 获取需要拼接的数据，报名分项，报名记录，报名分项记录 ==========================

            # ==================== 建立分项记录的字典映射  ==============================
            record_id_to_subitems_record_map = {}  # {"enroll_record_id":item}
            subitems_id_to_subitem_record_map = {}  # {"enroll_subitem_id":item}
            for enroll_subitems_record in enroll_subitems_record_list:
                # 判断当前用户是否存在报名分项记录
                if enroll_subitems_record["user_id"] or not enroll_subitems_record["user_id"] == user_id:
                    enroll_detail["this_user_has_subitem_record"] = 1
                # 找到属于其对应的报名记录, 并存如对应的字典中
                if record_id_to_subitems_record_map.get(enroll_subitems_record["enroll_subitem_id"], None) is None:
                    subitems_id_to_subitem_record_map[enroll_subitems_record["enroll_subitem_id"]] = [enroll_subitems_record]
                else:
                    subitems_id_to_subitem_record_map[enroll_subitems_record["enroll_subitem_id"]].append(enroll_subitems_record)

                if subitems_id_to_subitem_record_map.get(enroll_subitems_record["enroll_record_id"], None) is None:
                    record_id_to_subitems_record_map[enroll_subitems_record["enroll_record_id"]] = [enroll_subitems_record]
                else:
                    record_id_to_subitems_record_map[enroll_subitems_record["enroll_record_id"]].append(enroll_subitems_record)
            # ==================== 建立分享记录映射 ==============================

            # ==================== 报名记录和报名分项记录 进行数据拼接 ==========================
            for main_record in main_record_list:
                main_record["subitem_record_list"] = record_id_to_subitems_record_map.get(main_record["id"], {})
                # 遍历判断 该用户是否有报名记录
                if user_id and main_record["user_id"] and main_record["user_id"] == user_id:
                    enroll_detail["this_user_has_record"] = 1

            for enroll_subitem in enroll_subitems_list:
                enroll_subitem["subitem_record_list"] = subitems_id_to_subitem_record_map.get(enroll_subitem["id"], {})

            # 进行赋值
            enroll_detail["subitem_list"] = enroll_subitems_list
            enroll_detail["main_record_list"] = main_record_list
            # ==================== 报名记录和报名分项记录 进行数据拼接 ==========================

        return enroll_detail, None

    @staticmethod
    def enroll_list(params: dict = None, filter_fields: "list|str" = None, need_pagination: bool = True, **kwargs):
        """
        报名列表接口
        :param params: 查询参数
        :param filter_fields: 过滤字段列表或者字符串
        :param need_pagination: 是否需要分页
        :return: {'total': total, "page": page, "size": size, 'list': list(enroll_obj.object_list)}, err
        """
        # 字段处理
        kwargs, is_pass = force_transform_type(variable=kwargs, var_type="dict", default={})
        params, is_pass = force_transform_type(variable=params, var_type="dict", default={})
        params.update(kwargs)
        size, is_pass = force_transform_type(variable=params.pop('size', 10), var_type="int", default=10)
        page, is_pass = force_transform_type(variable=params.pop('page', 1), var_type="int", default=1)
        # 处理filter_fields，获取ORM查询字段列表
        filter_fields_list = filter_fields_handler(
            input_field_expression=filter_fields,
            all_field_list=EnrollServices.enroll_fields
        )
        filter_fields_list = list(set(filter_fields_list + ["thread_id", "user_id", "id"]))
        # 非法字段剔除，类型转换
        try:
            params = format_params_handle(
                param_dict=params,
                is_validate_type=True,
                filter_filed_list=[
                    "id|int", "id_list|list", "thread_id|int", "category_id|int", "category_id_list|list", "thread_id_list|list",
                    "user_id|int", "trading_relate", "region_code", "spend_time_start|date",
                    "enroll_status_code", "enroll_status_code_list|list",
                    "spend_time_end|date", "create_time_start|date", "create_time_end|date", "finish_time_start|date", "finish_time_end|date",
                    "open_time_start|date", "open_time_end|date", "enroll_status_code|int", "has_subitem|bool", "finance_invoicing_code"
                ],
                split_list=["category_id_list", "id_list", "enroll_status_code_list", "thread_id_list"],
                alias_dict={
                    "spend_time_start": "spend_time__gte", "spend_time_end": "spend_time__lte",
                    "create_time_start": "create_time__gte", "create_time_end": "create_time__lte",
                    "finish_time_start": "finish_time__gte", "open_time_start": "open_time__gte",
                    "open_time_end": "open_time__lte", "enroll_classify_value": "classify__value",
                    "enroll_category_value": "category__value",
                    "thread_id_list": "thread_id__in", "category_id_list": "category_id__in", "id_list": "id__in",
                    "enroll_status_code_list": "enroll_status_code__in"
                },
            )
        except ValueError as e:
            return None, str(e)
        # 构建ORM
        enroll_obj = Enroll.objects.extra(select={
            'register_time': 'DATE_FORMAT(register_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            'open_time': 'DATE_FORMAT(open_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            'close_time': 'DATE_FORMAT(close_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            'launch_time': 'DATE_FORMAT(launch_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            'finish_time': 'DATE_FORMAT(finish_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            'spend_time': 'DATE_FORMAT(spend_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            'create_time': 'DATE_FORMAT(create_time, "%%Y-%%m-%%d %%H:%%i:%%s")',
            'update_time': 'DATE_FORMAT(update_time, "%%Y-%%m-%%d %%H:%%i:%%s")'
        }).filter(**params).order_by("-id").values(*filter_fields_list)
        total = enroll_obj.count()

        #  # 分页展示, 满足不分页要求的时候则直接返回列表
        if not need_pagination and total <= 200:
            return list(enroll_obj), None
        else:
            paginator = Paginator(enroll_obj, size)
            try:
                enroll_obj = paginator.page(page)
            except EmptyPage:
                enroll_obj = paginator.page(paginator.num_pages)
            except Exception as e:
                return None, f'{str(e)}'
            return {'total': total, "page": page, "size": size, 'list': list(enroll_obj.object_list)}, None

    @staticmethod
    def enroll_own_list(params, need_pagination=True):
        """
        分页查看本人的报名列表
        :param params:
        :param need_pagination:
        :return:
        """
        page, is_pass = force_transform_type(variable=params.pop('page', 1), var_type="int", default=1)
        size, is_pass = force_transform_type(variable=params.pop('size', 10), var_type="int", default=10)

        params = format_params_handle(
            param_dict=params,
            filter_filed_list=[
                "enroll_id", "thread_id", "user_id", "enroll_pay_status", "enroll_status_code", "enroll_status_code_list",
                "price", "count", "amount", "category_id", "thread_id_list", "finance_invoicing_code"
            ],
            split_list=["enroll_status_code_list"],
            alias_dict={"enroll_id": "id", "enroll_status_code_list": "enroll_status_code__in", "thread_id_list": "thread_id__in"}
        )
        enroll_obj = Enroll.objects.filter(**params).order_by("-id").values()
        if not need_pagination:
            # serializer = EnrollRecordListSerializer(data=enroll_obj)
            # serializer.is_valid()
            return list(enroll_obj.object_list), None

        paginator = Paginator(enroll_obj, size)
        try:
            enroll_obj = paginator.page(page)
        except EmptyPage:
            enroll_obj = paginator.page(paginator.num_pages)
        except Exception as e:
            return None, f'{str(e)}'
        return {'total': paginator.count, "page": int(page), "size": int(size), 'list': list(enroll_obj.object_list)}, None

    @staticmethod
    def enroll_undertake_list(params, need_pagination=True):
        size = params.pop('size', 10)
        page = params.pop('page', 1)

        params = format_params_handle(
            param_dict=params,
            filter_filed_list=[
                "enroll_id", "category_id", "thread_id_list", "user_id", "enroll_pay_status", "enroll_status_code",
                "enroll_status_code_list", "price", "count", "amount", "finance_invoicing_code"
            ],
            split_list=["enroll_status_code_list"],
            alias_dict={
                "enroll_status_code_list": "enroll_status_code__in",
                "thread_id_list": "enroll__thread_id__in",
                "category_id": "enroll__category_id"
            }
        )
        enroll_obj = EnrollRecord.objects.annotate(
            commision=F("enroll__commision"),
            thread_id=F("enroll__thread_id")
        ).filter(**params).order_by("-id").values()

        if not need_pagination:
            # 扩展字段替换，并过滤
            return EnrollRecordServices.extend_transform_fields(list(enroll_obj.object_list)), None

        paginator = Paginator(enroll_obj, size)
        try:
            enroll_obj = paginator.page(page)
        except EmptyPage:
            enroll_obj = paginator.page(paginator.num_pages)
        except Exception as e:
            return None, f'{str(e)}'

        # 扩展字段替换，并过滤
        enroll_record_list = EnrollRecordServices.extend_transform_fields(list(enroll_obj.object_list))
        return {'total': paginator.count, "page": page, "size": size, 'list': enroll_record_list}, None

    @staticmethod
    def bxtx_pay_call_back(order_no):
        """
        镖行天下支付业务回调服务
        回调逻辑：
        1.用户支付分为两种：（1）用户支付预付款 （2）用户补差价
        情况1：支付预付款，但是还没有成单。修改状态进入代报名状态。
        情况2：（报名并指派后）用户支付差价，订单开始。进入上传状态。
        扩展逻辑：分销逻辑，资金逻辑。

        2023-2-6 回调逻辑：第一种代补差价付款成功后，直接进入已接单待上传，第二种：付完首付款后直接进入报名中
        :param order_no: 订单号
        :return: response
        """
        try:
            finance_transact_data, err = FinanceTransactsService.detail(order_no=order_no)
            write_to_log(
                prefix="报名支付回调",
                content="finance_transact_data:" + str(finance_transact_data or "") + "  err:" + str(err or "")
            )

            if err:
                raise Exception(err)

            user_id = finance_transact_data.get("account_id", None)
            enroll_id = finance_transact_data.get("enroll_id", None)
            outgo = finance_transact_data.get("outgo", 0)
            enroll_obj = Enroll.objects.filter(id=enroll_id)
            enroll_set = enroll_obj.first()
            if not enroll_set:
                return None, None
            paid_amount = abs(float(enroll_set.paid_amount or 0)) + abs(float(outgo or 0))
            if int(enroll_set.enroll_status_code) == 243:
                enroll_obj.update(paid_amount=paid_amount, unpaid_amount=0, enroll_status_code=356)
                EnrollSubitem.objects.filter(enroll_id=enroll_id).update(enroll_subitem_status_code=356)
                enroll_status_code = 356
            else:
                enroll_obj.update(paid_amount=paid_amount, unpaid_amount=0, enroll_status_code=232)
                EnrollSubitem.objects.filter(enroll_id=enroll_id).update(enroll_subitem_status_code=232)
                enroll_status_code = 232

            # 写入资金沙盒
            pass

            return enroll_status_code, None
        except Exception as e:
            write_to_log(prefix="报名支付回调异常", content="order_no:" + str(order_no or ""), err_obj=e)
            return None, str(e)

    @staticmethod
    def enroll_check_and_accept(enroll_id=None, check_success_code=656, **kwargs):
        """
        报名余额验收中
        :param enroll_id: 报名ID
        :param check_success_code: 当前审核中的状态码
        :param kwargs:  **
        :return:data,err
        """
        if not enroll_id or check_success_code is None:
            return None, "参数错误"
        is_check_success = Enroll.objects.filter(id=enroll_id, enroll_status_code=check_success_code).values("id").first()
        if not is_check_success:
            return None, "该报名不可触发余额"
        try:
            # ============ 完成订单联动资金修改 start ============
            # TODO 资金联动代码块，后期使用流程控制
            records_vales = list(
                EnrollRecord.objects.filter(enroll_id=enroll_id, enroll_status_code=check_success_code).values(
                    "user_id", "price", 'again_price', "initiator_again_price"
                )
            )
            for item in records_vales:
                content = {
                    "account_id": item.get("user_id"),
                    "amount": item.get("again_price", 0),
                    "currency": "CNY",
                    "pay_mode": "BALANCE",
                    "enroll_id": enroll_id
                }
                write_to_log(prefix="资金联动修改：", content=content)
                data, err = FinanceTransactsService.finance_create_or_write_off(data=content)
                if err:
                    write_to_log(level="error", prefix="资金余额联动修改异常，finance_create_or_write_off方法err", content=err)
            # ============ 完成订单联动资金修改 end ============
            return None, None
        except Exception as e:
            write_to_log(prefix="资金余额联动修改【enroll_check_and_accept】异常：", err_obj=e)
            return None, str(e)

    @staticmethod
    def old_enroll_check_and_accept(enroll_id=None, **kwargs):
        """
        报名余额验收中
        :param enroll_id: 报名ID
        :param kwargs:  **
        :return:data,err
        """
        if not enroll_id:
            return None, "参数错误，enroll_id为空"
        try:
            records_vales = list(
                EnrollRecord.objects.filter(enroll_id=enroll_id).exclude(enroll_status_code=124).values(
                    "user_id", "price", 'again_price', "initiator_again_price"
                )
            )
            for item in records_vales:
                content = {
                    "account_id": item.get("user_id"),
                    "amount": item.get("again_price", 0),
                    "currency": "CNY",
                    "pay_mode": "BALANCE",
                    "enroll_id": enroll_id
                }
                write_to_log(prefix="资金联动修改内容", content=content)
                data, err = FinanceTransactsService.finance_create_or_write_off(data=content)
                if err:
                    write_to_log(level="error", prefix="资金余额联动修改异常，finance_create_or_write_off方法err", content=err)
            return None, None
        except Exception as e:
            write_to_log(prefix="资金联动修改异常", err_obj=e)
            return None, str(e)

    # 已经弃用，当前业务逻辑改成指派机制，不在自动报满
    # @staticmethod
    # def enroll(request_params=None):
    #     """
    #     报名记录联合添加报名记录和分项记录
    #     :param request_params:
    #     :return:
    #     """
    #     sid = transaction.savepoint()
    #     # 信息主表
    #     main_record_params = format_params_handle(
    #         param_dict=request_params,
    #         filter_filed_list=[
    #             "id", "enroll_id", "user_id", "enroll_auth_status_id", "enroll_pay_status_id", "enroll_status_code",
    #             "create_time", "price", "deposit", "count",
    #             "main_amount", "coupon_amount", "again_reduction", "subitems_amount", "deposit_amount", "amount",
    #             "paid_amount", "unpaid_amount", "fee", "photos",
    #             "files", "score", "reply", "remark",
    #         ]
    #     )
    #     main_data, err = EnrollRecordServices.record_add(main_record_params)
    #     if err:
    #         transaction.savepoint_rollback(sid)
    #         return None, err
    #
    #     # 分项表
    #     subitem_params = format_params_handle(
    #         param_dict=request_params,
    #         filter_filed_list=["enroll_record_id", "enroll_subitem_id", "user_id", "subitem_price", "subitem_count",
    #                            "subitem_amount", ],
    #         alias_dict={"subitem_price": "price", "subitem_count": "count"}
    #     )
    #
    #     enroll_record_id = main_data.get("id", None)
    #     if not enroll_record_id:
    #         transaction.savepoint_rollback(sid)
    #         return None, err
    #     subitem_params["enroll_record_id"] = enroll_record_id
    #
    #     subitem_data, err = EnrollSubitemRecordService.add(subitem_params)
    #     if err:
    #         transaction.savepoint_rollback(sid)
    #         return None, err
    #
    #     # 添加定时器
    #     clocker = ClockService()
    #     clocker.add_clock(enroll_id=request_params.get("enroll_id"), user_id=request_params.get("user_id"))
    #     transaction.clean_savepoints()
    #     return None, err



"""
Created on 2022-05-19
@author:刘飞
@description:报名模块序列化器
"""
from rest_framework import serializers

from .models import *


class EnrollListSerializer(serializers.ModelSerializer):
    """
    报名表序列化器
    """

    class Meta:
        model = Enroll
        fields = '__all__'


class EnrollRecordListSerializer(serializers.ModelSerializer):
    """
    报名记录列表序列化器
    """
    price = serializers.DecimalField(max_digits=10, decimal_places=2)
    deposit = serializers.DecimalField(max_digits=10, decimal_places=2)
    count = serializers.IntegerField()
    main_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    coupon_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    again_reduction = serializers.DecimalField(max_digits=10, decimal_places=2)
    subitems_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    deposit_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    paid_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    unpaid_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    fee = serializers.DecimalField(max_digits=10, decimal_places=2)
    create_time = serializers.DateTimeField(format='%Y-%m-%d %H:%M:%S')
    thread_id = serializers.IntegerField()

    class Meta:
        model = EnrollRecord
        fields = '__all__'

class EnrollRecordListV2Serializer(serializers.ModelSerializer):
    """
    报名记录列表序列化器
    """
    price = serializers.DecimalField(max_digits=10, decimal_places=2)
    deposit = serializers.DecimalField(max_digits=10, decimal_places=2)
    count = serializers.IntegerField()
    main_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    coupon_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    again_reduction = serializers.DecimalField(max_digits=10, decimal_places=2)
    subitems_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    deposit_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    paid_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    unpaid_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    fee = serializers.DecimalField(max_digits=10, decimal_places=2)
    create_time = serializers.DateTimeField(format='%Y-%m-%d %H:%M:%S')
    thread_id = serializers.IntegerField()
    enroll_user_id = serializers.IntegerField()

    class Meta:
        model = EnrollRecord
        fields = '__all__'


class EnrollSubitemRecordSerializer(serializers.ModelSerializer):
    """
    报名记录列表序列化器
    """
    price = serializers.DecimalField(max_digits=10, decimal_places=2)
    subitem_amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    count = serializers.IntegerField()

    class Meta:
        model = EnrollSubitemRecord
        fields = '__all__'


class EnrollSubitemSerializer(serializers.ModelSerializer):
    """
    报名记录列表序列化器
    """
    price = serializers.DecimalField(max_digits=10, decimal_places=2)
    amount = serializers.DecimalField(max_digits=10, decimal_places=2)
    count = serializers.IntegerField()

    class Meta:
        model = EnrollSubitem
        fields = '__all__'



# encoding: utf-8
"""
@project: djangoModel->tool
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis: 返回协议封装工具
@created_time: 2022/6/15 14:14
"""
import json

from django.http import JsonResponse


# json 结果集返回
def parse_json(result):
    if not result is None:
        if type(result) is str:
            try:
                result = json.loads(result.replace("'", '"').replace('\\r', "").replace('\\n', "").replace('\\t', "").replace('\\t', ""))
            except Exception as e:
                return result
        if type(result) is list or type(result) is tuple:
            for index, value in enumerate(result):
                result[index] = parse_json(value)
        if type(result) is dict:
            for k, v in result.items():
                result[k] = parse_json(v)
    return result


# 数据返回规则
def util_response(data='', err=0, msg='ok', is_need_parse_json=True):
    """
    http 返回协议封装
    :param data: 返回的数据体
    :param err: 错误码，一般以1000开始，逐一增加。登录错误为6000-6999。
    :param msg: 错误信息，一般为服务返回协议中的err,自动解析内容
    :return: response对象
    """
    data = parse_json(data) if is_need_parse_json else data
    response_json = {'err': err, 'data': data}
    # 解析msg字符串
    try:
        msg_list = msg.split(";")
        if len(msg_list) <= 1:
            response_json["msg"] = msg
        else:
            for i in msg_list:
                [key, value] = i.split(":")
                response_json[key] = value
    except Exception as e:
        response_json["msg"] = msg

    return JsonResponse(response_json)



# encoding: utf-8
"""
@project: djangoModel->j_valuation
@author: 孙楷炎
@Email: sky4834@163.com
@synopsis: 计价服务类
@created_time: 2022/10/13 16:58
"""
import math
import re

from .custom_tool import write_to_log


# 计算器基类
class JBaseExpression:
    """
    公式计算积累
    """
    __result = None  # 结果
    __expression = None
    fun_type = ""

    @staticmethod
    def get_child_info():
        """
        获取所有子类,返回公式类的名称与类指针的映射
        :return: {"IF":IFExpression}
        """
        try:
            return {getattr(i, "name"): i for i in JBaseExpression.__subclasses__() if not getattr(i, "name", None) is None}, None
        except AttributeError:
            return {}, "扩展类name属性不可以重复"

    # 解析变量
    @staticmethod
    def parse_variables(expression_string, input_dict):
        """
        解析包含变量的计算公式字符串,
        从 input_dict 变量字典中拿到对应变量进行替换，没有则取0。
        :param expression_string: 带有未处理的变量公式字符串： {{price}} * {{count}} - {{reduction}} + {{fee}}
        :param input_dict: 变量字典如： {'count': 1, 'price': 50, 'subitem__amount': '5,6,8', 'subitem__count': '100'}
        :return: "50 * 1 - 0 + 0 + 19 + sum(0)", err
        """
        try:
            if expression_string is None:
                return "0", None

            # 在 expression_string 中提取出变量
            this_cell_value_match = re.compile("{{.*?}}").findall(str(expression_string))

            # 把提取出的变量与值进行映射, {"{{price}}":100}
            parsed_variable_map = {}
            for i in this_cell_value_match:
                parsed_variable_map.update({i: input_dict.get(i.replace("{{", "").replace("}}", ""), "0")})

            # 根据映射对公式进行替换, {{price}}替换为 100
            for k, v in parsed_variable_map.items():
                expression_string = expression_string.replace(k, str(v) if v else "0")

            return expression_string, None

        except Exception as e:
            write_to_log(prefix="计价解析变量异常", err_obj=e)
            return "0", str(e)

    # 解析括号，成对解析
    @staticmethod
    def parsed_brackets(expression, need_list=False):
        """
        递归解析变量中的括号，并调用计价进行计算
        :param expression: 不含有变量的公式字符串,如：50 * 1 - 0 + 0 + 19 + sum(0)
        :param need_list: 是否需要
        :return: expression中括号的索引元组如：[(start_index,end_index)]
        """
        twain_index_map = {}
        twain_index_list = []
        forward_index_list = []
        for char, index in zip(expression, range(len(expression))):
            if char == "(":
                forward_index_list.append(index)
            if char == ")":
                forward_bracket = forward_index_list.pop(-1)
                twain_index_map[forward_bracket] = (forward_bracket, index)
                twain_index_list.append((forward_bracket, index))
        return twain_index_list if need_list else twain_index_map

    # 检测是否可直接解析的公式，是否存在公式。
    def has_expression(self, expression_string):
        """
        判断expression_string中是否存在未计算的公式
        :param expression_string:
        :return: True/False
        """
        child_class, err = self.get_child_info()
        fun_patt = ""
        for name, instance in child_class.items():
            if not name:
                continue
            fun_patt = "(" + name + ")" if fun_patt == "" else fun_patt + "|" + "(" + name + ")"

        fun_patt = "(" + fun_patt + ")"  # 自定义公名称匹配：((IF)|(SUM)|(CEIL))
        return True if re.search(fun_patt, expression_string) else False

    # 解析公式 复杂度比较高 有优化的空间
    def parse_expression(self, expression_string):
        # 解析括号和函数
        # 解析函数
        replace_list = []
        child_class, err = self.get_child_info()
        for name, instance in child_class.items():
            # 基础计算公式跳过
            if not name:
                continue
            for r in re.finditer(instance.patt, expression_string):
                expression_start, expression_end = r.span()  # 找到未解析德公式
                full_fun_string = expression_string[expression_start:expression_end]
                fun_inner_string = full_fun_string.replace(name, "")
                # 内部存在公式不计算，方式无限递归
                if not self.has_expression(fun_inner_string):
                    result = instance().process(full_fun_string)
                    replace_list.append((full_fun_string, result))

        # 最内层结果替换
        for full_fun_string, result in replace_list:
            expression_string = expression_string.replace(full_fun_string, str(result))

        # 如果还有没有处理的公式，则递归解析
        if self.has_expression(expression_string):  # 递归解析
            expression_string = self.parse_expression(expression_string)
        return expression_string


# ================================= 公式封装类 start =======================================
class CalculateExpression(JBaseExpression):
    name = ""
    patt = "\([^\(]*?\)"

    def process(self, expression_string):
        try:
            return eval(expression_string)  # 加加减乘除
        except ZeroDivisionError:
            return 0  # 0/n  的情况
        except Exception as e:
            write_to_log(level="error", prefix="计价异常（SUMExpression）", err_obj=e)
            return expression_string


class IFExpression(JBaseExpression):
    """
    布尔值判断、三元运算
    """
    name = "IF"
    patt = "IF\([^\(]*?\)"

    def process(self, expression, *args, ):
        bool_expression, yes_expression, no_expression = expression.replace(IFExpression.name, "").replace("(", "").replace(")", "").split(",")
        if CalculateExpression().process(bool_expression):
            return yes_expression
        else:
            return no_expression


class SUMExpression(JBaseExpression):
    """
    求和计算、（多参数在服务中获取）
    """
    name = "SUM"
    patt = "SUM\([^\(]*?\)"  # SUM(a,b,c) == a+b+c

    def process(self, expression):
        try:
            args = expression.replace(self.name, "").replace("(", "").replace(")", "").split(",")
            result = 0
            for i in args:
                result = eval(str(result) + "+" + str(i))
            return result
        except Exception as e:
            write_to_log(level="error", prefix="计价异常（SUMExpression）", err_obj=e)
            return 0


class CEILExpression(JBaseExpression):
    """
    取整公式
    """
    name = "CEIL"
    patt = "CEIL\([^\(]*?\)"  # ROUND(num, index, step)

    def process(self, expression):
        try:
            expression = expression.replace(self.name, "").replace("(", "").replace(")", "")
            return math.ceil(eval(expression))
        except Exception as e:
            write_to_log(level="error", prefix="计价异常（SUMExpression）", err_obj=e)
            return 0


# ================================= 公式封装类 end  =======================================

# ================================= 算入口类，计算器 对外暴露的类 start =======================================

class JExpression(JBaseExpression):
    """
    计算入口类，计算器 对外暴露的类
    """

    def process(self, expression_string):
        # 公式合法性校验：expression_string
        if expression_string is None:
            return "", None
        if not len(re.findall("\(", expression_string)) == len(re.findall("\)", expression_string)):
            write_to_log(prefix="计价公式配置错误，请及时检查", content="语法错误，括号应该成对存在")
            return None, "语法错误，括号应该成对存在"
        # 公式字符串,符号预处里预处理
        expression_string = expression_string.upper().replace(" ", "").replace("=", "==").replace(">==", ">=").replace("<==", "<=")

        # 把自定义公式计算出结果进行替换，得出基础的加减乘除公式字符拆
        parsed_expression_string = self.parse_expression(expression_string)

        # 计算基础加减乘除，然后得出结果并返回
        result = CalculateExpression().process(parsed_expression_string)
        return result, None
# ================================= 算入口类，计算器 对外暴露的类 end   =======================================
